from random import randint, shuffle
from typing import List, Tuple

import pytest
import torch
from torch import Tensor

TLabelsLI = List[Tuple[List[int], int, int]]


def generate_valid_labels(num: int) -> TLabelsLI:
    """
    This function generates some valid inputs for miners.
    It generates n_instances for n_labels.

    Args:
        num: Number of generated samples

    Returns:
        Samples in the following order: (labels, n_labels, n_instances)

    """
    labels_li = []

    for _ in range(num):
        n_labels, n_instances = randint(2, 12), randint(2, 12)
        labels_list = [[label] * randint(2, 12) for label in range(n_labels)]
        labels = [el for sublist in labels_list for el in sublist]

        shuffle(labels)
        labels_li.append((labels, n_labels, n_instances))

    return labels_li


@pytest.fixture()
def features_and_labels() -> List[Tuple[Tensor, List[int]]]:
    """
    Returns: list of features and valid labels

    """
    num_batches = 100
    features_dim = 10

    labels_li = generate_valid_labels(num=num_batches)
    labels_list, _, _ = zip(*labels_li)

    features = []
    for labels in labels_list:
        features.append(torch.rand(size=(len(labels), features_dim)))

    return list(zip(features, labels_list))



from oml.models.utils import remove_prefix_from_state_dict


def test_remove_prefix_from_state_dict() -> None:
    inp = {"model.layer4": 1}, "layer4"
    gt = {"layer4": 1}
    ans = remove_prefix_from_state_dict(*inp)  # type: ignore

    assert gt == ans, ans

    inp = {"layer4": 1}, "layer4"
    gt = {"layer4": 1}
    ans = remove_prefix_from_state_dict(*inp)  # type: ignore

    assert gt == ans, ans



from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union

import albumentations as albu
import numpy as np
import pandas as pd
import torchvision
from torch.utils.data import Dataset

from oml.const import (
    CATEGORIES_COLUMN,
    CATEGORIES_KEY,
    INPUT_TENSORS_KEY,
    IS_GALLERY_COLUMN,
    IS_GALLERY_KEY,
    IS_QUERY_COLUMN,
    IS_QUERY_KEY,
    LABELS_COLUMN,
    LABELS_KEY,
    PATHS_COLUMN,
    PATHS_KEY,
    SPLIT_COLUMN,
    X1_COLUMN,
    X1_KEY,
    X2_COLUMN,
    X2_KEY,
    Y1_COLUMN,
    Y1_KEY,
    Y2_COLUMN,
    Y2_KEY,
)
from oml.interfaces.datasets import IDatasetQueryGallery, IDatasetWithLabels
from oml.registry.transforms import get_transforms
from oml.transforms.images.utils import TTransforms, get_im_reader_for_transforms
from oml.utils.dataframe_format import check_retrieval_dataframe_format
from oml.utils.images.images import TImReader, imread_cv2


class BaseDataset(Dataset):
    """
    Base class for the retrieval datasets.

    """

    def __init__(
        self,
        df: pd.DataFrame,
        transform: Optional[TTransforms] = None,
        dataset_root: Optional[Union[str, Path]] = None,
        f_imread: TImReader = imread_cv2,
        cache_size: int = 100_000,
        input_tensors_key: str = INPUT_TENSORS_KEY,
        labels_key: str = LABELS_KEY,
        paths_key: str = PATHS_KEY,
        categories_key: Optional[str] = CATEGORIES_KEY,
        x1_key: str = X1_KEY,
        x2_key: str = X2_KEY,
        y1_key: str = Y1_KEY,
        y2_key: str = Y2_KEY,
    ):
        """

        Args:
            df: Table with the following obligatory columns:

                  >>> LABELS_COLUMN, PATHS_COLUMN

                  and the optional ones:

                  >>> X1_COLUMN, X2_COLUMN, Y1_COLUMN, Y2_COLUMN, CATEGORIES_COLUMN

            transform: Augmentations for the images, set ``None`` to perform only normalisation and casting to tensor
            dataset_root: Path to the images dir, set ``None`` if you provided the absolute paths in your dataframe
            f_imread: Function to read the images
            cache_size: Size of the dataset's cache
            input_tensors_key: Key to put tensors into the batches
            labels_key: Key to put labels into the batches
            paths_key: Key put paths into the batches
            categories_key: Key to put categories into the batches
            x1_key: Key to put ``x1`` into the batches
            x2_key: Key to put ``x2`` into the batches
            y1_key: Key to put ``y1`` into the batches
            y2_key: Key to put ``y2`` into the batches

        """
        df = df.copy()

        assert all(x in df.columns for x in (LABELS_COLUMN, PATHS_COLUMN))

        self.input_tensors_key = input_tensors_key
        self.labels_key = labels_key
        self.paths_key = paths_key
        self.categories_key = categories_key if (CATEGORIES_COLUMN in df.columns) else None

        self.bboxes_exist = all(coord in df.columns for coord in (X1_COLUMN, X2_COLUMN, Y1_COLUMN, Y2_COLUMN))
        if self.bboxes_exist:
            self.x1_key, self.x2_key, self.y1_key, self.y2_key = x1_key, x2_key, y1_key, y2_key
        else:
            self.x1_key, self.x2_key, self.y1_key, self.y2_key = None, None, None, None

        if dataset_root is not None:
            dataset_root = Path(dataset_root)
            df[PATHS_COLUMN] = df[PATHS_COLUMN].apply(lambda x: str(dataset_root / x))
        else:
            df[PATHS_COLUMN] = df[PATHS_COLUMN].astype(str)

        self.df = df
        self.transform = transform if transform else get_transforms("norm_albu")
        self.f_imread = f_imread
        self.read_bytes_image_cached = lru_cache(maxsize=cache_size)(self._read_bytes_image)

        available_augs_types = (albu.Compose, torchvision.transforms.Compose)
        assert isinstance(self.transform, available_augs_types), f"Type of transforms must be in {available_augs_types}"

    @staticmethod
    def _read_bytes_image(path: Union[Path, str]) -> bytes:
        with open(str(path), "rb") as fin:
            return fin.read()

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        row = self.df.iloc[idx]

        img_bytes = self.read_bytes_image_cached(row[PATHS_COLUMN])
        img = self.f_imread(img_bytes)

        im_h, im_w = img.shape[:2] if isinstance(img, np.ndarray) else img.size[::-1]

        if (not self.bboxes_exist) or any(
            pd.isna(coord) for coord in [row[X1_COLUMN], row[X2_COLUMN], row[Y1_COLUMN], row[Y2_COLUMN]]
        ):
            x1, y1, x2, y2 = 0, 0, im_w, im_h
        else:
            x1, y1, x2, y2 = int(row[X1_COLUMN]), int(row[Y1_COLUMN]), int(row[X2_COLUMN]), int(row[Y2_COLUMN])

        if isinstance(self.transform, albu.Compose):
            img = img[y1:y2, x1:x2, :]  # todo: since albu may handle bboxes we should move it to augs
            image_tensor = self.transform(image=img)["image"]
        else:
            # torchvision.transforms
            img = img.crop((x1, y1, x2, y2))
            image_tensor = self.transform(img)

        item = {
            self.input_tensors_key: image_tensor,
            self.labels_key: row[LABELS_COLUMN],
            self.paths_key: row[PATHS_COLUMN],
        }

        if self.categories_key:
            item[self.categories_key] = row[CATEGORIES_COLUMN]

        if self.bboxes_exist:
            item.update(
                {
                    self.x1_key: x1,
                    self.y1_key: y1,
                    self.x2_key: x2,
                    self.y2_key: y2,
                }
            )

        return item

    def __len__(self) -> int:
        return len(self.df)

    @property
    def bboxes_keys(self) -> Tuple[str, ...]:
        if self.bboxes_exist:
            return self.x1_key, self.y1_key, self.x2_key, self.y2_key
        else:
            return tuple()


class DatasetWithLabels(BaseDataset, IDatasetWithLabels):
    """
    The main purpose of this class is to be used as a dataset during
    the training stage.

    It has to know how to return its labels, which is required information
    to perform the training with the combinations-based losses.
    Particularly, these labels will be passed to `Sampler` to form the batches and
    batches will be passed to `Miner` to form the combinations (triplets).

    """

    def get_labels(self) -> np.ndarray:
        return np.array(self.df[LABELS_COLUMN].tolist())


class DatasetQueryGallery(BaseDataset, IDatasetQueryGallery):
    """
    The main purpose of this class is to be used as a dataset during
    the validation stage. It has to provide information
    about its `query`/`gallery` split.

    Note, that some of the datasets used as benchmarks in Metric Learning
    provide the splitting information (for example, ``DeepFashion InShop`` dataset), but some of them
    don't (for example, ``CARS196`` or ``CUB200``).
    The validation idea for the latter is to calculate the embeddings for the whole validation set,
    then for every item find ``top-k`` nearest neighbors and calculate the desired retrieval metric.
    In other words, for the desired query item, the gallery is the rest of the validation dataset.

    Thus, if you want to perform this kind of validation process (`1 vs rest`) you should simply return
    ``is_query == True`` and ``is_gallery == True`` for every item in the dataset sa the same time.

    """

    def __init__(
        self,
        df: pd.DataFrame,
        dataset_root: Optional[Union[str, Path]] = None,
        transform: Optional[albu.Compose] = None,
        f_imread: TImReader = imread_cv2,
        cache_size: int = 100_000,
        input_tensors_key: str = INPUT_TENSORS_KEY,
        labels_key: str = LABELS_KEY,
        paths_key: str = PATHS_KEY,
        categories_key: str = CATEGORIES_KEY,
        x1_key: str = X1_KEY,
        x2_key: str = X2_KEY,
        y1_key: str = Y1_KEY,
        y2_key: str = Y2_KEY,
        is_query_key: str = IS_QUERY_KEY,
        is_gallery_key: str = IS_GALLERY_KEY,
    ):
        super(DatasetQueryGallery, self).__init__(
            df=df,
            dataset_root=dataset_root,
            transform=transform,
            f_imread=f_imread,
            cache_size=cache_size,
            input_tensors_key=input_tensors_key,
            labels_key=labels_key,
            paths_key=paths_key,
            categories_key=categories_key,
            x1_key=x1_key,
            x2_key=x2_key,
            y1_key=y1_key,
            y2_key=y2_key,
        )
        assert all(x in df.columns for x in (IS_QUERY_COLUMN, IS_GALLERY_COLUMN))

        self.is_query_key = is_query_key
        self.is_gallery_key = is_gallery_key

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        item = super().__getitem__(idx)
        item[self.is_query_key] = bool(self.df.iloc[idx][IS_QUERY_COLUMN])
        item[self.is_gallery_key] = bool(self.df.iloc[idx][IS_GALLERY_COLUMN])
        return item


def get_retrieval_datasets(
    dataset_root: Path,
    transforms_train: Any,
    transforms_val: Any,
    f_imread_train: Optional[TImReader] = None,
    f_imread_val: Optional[TImReader] = None,
    dataframe_name: str = "df.csv",
    cache_size: int = 100_000,
) -> Tuple[DatasetWithLabels, DatasetQueryGallery]:
    df = pd.read_csv(dataset_root / dataframe_name, index_col=False)
    check_retrieval_dataframe_format(df, dataset_root=dataset_root)

    # train
    df_train = df[df[SPLIT_COLUMN] == "train"].reset_index(drop=True)
    train_dataset = DatasetWithLabels(
        df=df_train,
        dataset_root=dataset_root,
        transform=transforms_train,
        cache_size=cache_size,
        f_imread=f_imread_train or get_im_reader_for_transforms(transforms_train),
    )

    # val (query + gallery)
    df_query_gallery = df[df[SPLIT_COLUMN] == "validation"].reset_index(drop=True)
    valid_dataset = DatasetQueryGallery(
        df=df_query_gallery,
        dataset_root=dataset_root,
        transform=transforms_val,
        cache_size=cache_size,
        f_imread=f_imread_val or get_im_reader_for_transforms(transforms_val),
    )

    return train_dataset, valid_dataset


__all__ = ["BaseDataset", "DatasetWithLabels", "DatasetQueryGallery", "get_retrieval_datasets"]



import random
from itertools import chain
from typing import List, Tuple, Union

import torch

from oml.functional.metrics import calc_mask_to_ignore


def generate_retrieval_case(
    max_labels: int,
    max_samples_per_label: int,
    is_query_all: bool = False,
    is_gallery_all: bool = False,
    return_desired_correct_positions: bool = False,
) -> Tuple[Union[List[List[int]], torch.Tensor], torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Function generates set of parameters `is_label`, `is_query`, `is_gallery` as well as `distance_matrix` or
    `desired_correct_positions` of GT

    Args:
        max_labels: maximum number of generated labels (minimum is 2)
        max_samples_per_label: maximum number of generated samples per label (minimum is 2). Each unique label
            can have different number of samples
        is_query_all: Set each generated label as query
        is_gallery_all: Set each generated label as gallery
        return_desired_correct_positions: If False return distance matrix instead of GT positions matrix

    Example:
        max_labels = 4  # 2 - 4 unqiue labels
        max_samples_per_label = 4  # 2 - 4 samples per label

        =>

        labels =
        [0, 0,
        1, 1, 1, 1,
        2, 2, 2]
        is_query =
        [True, False,
        False, True, True, False,
        True, False, True]
        is_gallery =
        [True, True,
        True, False, True, False,
        False, True, True]
    """

    assert max_labels >= 2
    assert max_samples_per_label >= 2

    range_num_labels = (2, max_labels)
    range_num_samples_per_label = (2, max_samples_per_label)

    labels = []
    is_query = []
    is_gallery = []

    num_samples = [random.randint(*range_num_samples_per_label) for _ in range(random.randint(*range_num_labels))]

    for label, num_samples_for_label in enumerate(num_samples):
        labels.extend([label] * num_samples_for_label)

        if is_query_all:
            is_query_for_label = [True] * num_samples_for_label
        else:
            is_query_for_label = [bool(random.randint(0, 1)) for _ in range(num_samples_for_label)]
        guaranted_query = random.randrange(num_samples_for_label)
        is_query_for_label[guaranted_query] = True

        if is_gallery_all:
            is_gallery_for_label = [True] * num_samples_for_label
        else:
            is_gallery_for_label = [bool(random.randint(0, 1)) for _ in range(num_samples_for_label)]

            # we guarantee that for each query there is at least one gallery which is not equal to query
            if sum(is_query_for_label) == 1:
                # set another available idx as gallery
                guaranted_galleries = random.sample(set(range(num_samples_for_label)) - {guaranted_query}, k=1)
            else:
                # We have at least 2 queries in this section
                # To calculate metric for each query it's necessary to have minimum 1 gallery with another idx.
                # With k=2 one gallery can be from same or another idx, but second is guaranteed with another idx
                # E.g. is_query = [True, True, True] and is_gallery = [True, False, True]
                guaranted_galleries = random.sample(range(num_samples_for_label), k=2)

            for guaranted_g in guaranted_galleries:
                is_gallery_for_label[guaranted_g] = True

        is_query.append(is_query_for_label)
        is_gallery.append(is_gallery_for_label)

    num_galleries_total = sum(chain(*is_gallery))

    def _gen_correct_positions(_is_gallery_for_label: List[bool], q_is_g: bool) -> List[int]:
        return sorted(random.sample(range(num_galleries_total - q_is_g), k=sum(_is_gallery_for_label) - q_is_g))

    desired_correct_positions = []
    for is_query_for_label, is_gallery_for_label in zip(is_query, is_gallery):
        for q, g in zip(is_query_for_label, is_gallery_for_label):
            if q:
                desired_correct_positions.append(_gen_correct_positions(is_gallery_for_label, q == g))

    labels = torch.tensor(labels)
    is_query = torch.tensor(list(chain(*is_query)))
    is_gallery = torch.tensor(list(chain(*is_gallery)))

    if return_desired_correct_positions:
        return desired_correct_positions, labels, is_query, is_gallery
    else:
        distances = generate_distance_matrix(
            desired_correct_positions, labels=labels, is_query=is_query, is_gallery=is_gallery
        )
        return distances, labels, is_query, is_gallery


def generate_distance_matrix(
    correct_positions_array: List[List[int]], labels: torch.Tensor, is_query: torch.Tensor, is_gallery: torch.Tensor
) -> torch.Tensor:
    """
    Function generates distance matrix based on desired correct positions.
    For cases query == gallery coresponded distances are equal to zero

    Example:

        >>> desired_correct_positions = [[0, 1, 4], [0, 2, 4]]
        >>> labels = torch.tensor([0, 0, 0, 0, 1, 1])
        >>> is_query = torch.tensor([False, True, True, False, False, False])
        >>> is_gallery = torch.tensor([True] * len(labels))

        gives a distance matrix which, after sorting and applying `mask_to_ignore`, has the correct answers at
        the corresponding positions

        VVXXV
        VXVXV

        where V - correct label, X - incorrect label
    """
    num_query = is_query.sum()
    num_gallery = is_gallery.sum()

    assert num_query == len(correct_positions_array)

    mask_to_ignore = calc_mask_to_ignore(is_query=is_query, is_gallery=is_gallery)
    q_in_g_array = torch.any(mask_to_ignore, dim=1).tolist()
    query_ids = torch.nonzero(is_query).squeeze()

    assert all(len(pos) < num_gallery for pos in correct_positions_array)

    for idx, (pos, q_in_g) in enumerate(zip(correct_positions_array, q_in_g_array)):
        if not all(idx < num_gallery - int(q_in_g) for idx in pos):
            raise ValueError(
                f"Desired ids are greater than available number of galleries for "
                f"label_idx={query_ids[idx].item()}, desired_positions={pos}"
            )
        if not (len(pos) < (num_gallery - int(q_in_g))):
            raise ValueError(
                f"Desired number of ids are greater than available number of galleries for "
                f"label_idx={query_ids[idx].item()}, desired_positions={pos}"
            )

    gallery_ids = torch.nonzero(is_gallery).squeeze()
    gallery_labels = labels[is_gallery]

    distances = -1 * torch.ones(num_query, num_gallery, dtype=torch.float)

    for row, (query_idx, correct_positions, q_in_g) in enumerate(zip(query_ids, correct_positions_array, q_in_g_array)):
        same_labels = gallery_labels == labels[query_idx]
        idx_q_is_g_in_same_labels = torch.nonzero(gallery_ids == query_idx).squeeze()

        if q_in_g:
            same_labels[idx_q_is_g_in_same_labels] = False

        distances[row, same_labels] = torch.tensor(correct_positions).float()

        if q_in_g:
            same_labels[idx_q_is_g_in_same_labels] = True

        rest_of_distances = set(range(num_gallery - int(q_in_g))) - set(correct_positions)
        rest_of_distances = torch.tensor(sorted(rest_of_distances)).float()

        distances[row, ~same_labels] = rest_of_distances

    distances += 1
    distances[mask_to_ignore] = 0

    return distances



from functools import partial
from typing import Any

import matplotlib.pyplot as plt
import torch

from oml.const import (
    CATEGORIES_KEY,
    EMBEDDINGS_KEY,
    IS_GALLERY_KEY,
    IS_QUERY_KEY,
    LABELS_KEY,
    PATHS_KEY,
    get_cache_folder,
)
from oml.metrics.embeddings import EmbeddingMetrics
from oml.utils.misc import one_hot

oh = partial(one_hot, dim=8)


def test_visualization() -> None:

    dummy_image = [[0]]

    cf = get_cache_folder()
    plt.imsave(cf / "temp.png", dummy_image)

    batch1 = {
        EMBEDDINGS_KEY: torch.stack([oh(1) * 2, oh(1) * 3, oh(0)]),
        LABELS_KEY: torch.tensor([0, 1, 1]),
        IS_QUERY_KEY: torch.tensor([True, True, True]),
        IS_GALLERY_KEY: torch.tensor([False, False, False]),
        CATEGORIES_KEY: torch.tensor([10, 20, 20]),
        PATHS_KEY: [cf / "temp.png", cf / "temp.png", cf / "temp.png"],
    }

    batch2 = {
        EMBEDDINGS_KEY: torch.stack([oh(0), oh(1), oh(1)]),
        LABELS_KEY: torch.tensor([0, 1, 1]),
        IS_QUERY_KEY: torch.tensor([False, False, False]),
        IS_GALLERY_KEY: torch.tensor([True, True, True]),
        CATEGORIES_KEY: torch.tensor([10, 20, 20]),
        PATHS_KEY: [cf / "temp.png", cf / "temp.png", cf / "temp.png"],
    }

    calc = EmbeddingMetrics(
        embeddings_key=EMBEDDINGS_KEY,
        labels_key=LABELS_KEY,
        is_query_key=IS_QUERY_KEY,
        is_gallery_key=IS_GALLERY_KEY,
        extra_keys=(PATHS_KEY,),
        map_top_k=(2,),
    )

    num_samples = len(batch1[LABELS_KEY]) + len(batch2[LABELS_KEY])

    calc.setup(num_samples=num_samples)
    calc.update_data(batch1)
    calc.update_data(batch2)

    calc.compute_metrics()

    figs, log_strs = calc.visualize()

    (cf / "temp.png").unlink()

    assert len(figs) == len(log_strs)



from collections import defaultdict
from typing import List, Tuple

import numpy as np
import pytest
import torch

from oml.functional.metrics import (
    TMetricsDict,
    calc_gt_mask,
    calc_mask_to_ignore,
    calc_retrieval_metrics,
    validate_dataset,
)

from .synthetic import generate_distance_matrix, generate_retrieval_case

TPositions = List[List[int]]


def naive_cmc(positions: TPositions, k: int) -> torch.Tensor:
    values = torch.empty(len(positions), dtype=torch.bool)
    for query_idx, pos in enumerate(positions):
        values[query_idx] = any(idx < k for idx in pos)
    metric = torch.mean(values.float())
    return metric


def naive_map(positions: TPositions, k: int) -> torch.Tensor:
    values = torch.empty(len(positions), dtype=torch.float)
    for query_idx, pos in enumerate(positions):
        num_gt = min(len(pos), k)
        values[query_idx] = sum(gt_count / (idx + 1) for gt_count, idx in enumerate(pos, 1) if idx < k) / num_gt
    metric = torch.mean(values.float())
    return metric


def naive_precision(positions: TPositions, k: int) -> torch.Tensor:
    values = torch.empty(len(positions), dtype=torch.float)
    for query_idx, pos in enumerate(positions):
        num_gt = min(len(pos), k)
        values[query_idx] = sum(1 for idx in pos if idx < k) / num_gt
    metric = torch.mean(values.float())
    return metric


def test_on_exact_case() -> None:
    """
    label 0:
    VVXVX
    VXXVV
    XVVXV
    XXVVV
    label 1:
    VXXXX
    XXXVX
    """
    labels = torch.tensor([0, 0, 0, 0, 1, 1])
    is_query = torch.tensor([True] * len(labels))
    is_gallery = torch.tensor([True] * len(labels))
    positions = [[0, 1, 3], [0, 3, 4], [1, 2, 4], [2, 3, 4], [0], [3]]
    top_k = (1, 2, 3, 4, 5, 10)

    metrics_expected = defaultdict(dict)  # type: ignore

    metrics_expected["precision"][1] = torch.tensor([1, 1, 0, 0, 1, 0]).float()
    metrics_expected["map"][1] = torch.tensor([1, 1, 0, 0, 1, 0]).float()

    metrics_expected["precision"][2] = torch.tensor([1, 0.5, 0.5, 0, 1, 0]).float()
    metrics_expected["map"][2] = torch.tensor([1, 0.5, 0.25, 0, 1, 0]).float()

    metrics_expected["precision"][3] = torch.tensor([0.6666, 0.3333, 0.6666, 0.3333, 1, 0]).float()
    metrics_expected["map"][3] = torch.tensor([0.6666, 0.3333, 0.3888, 0.1111, 1, 0]).float()

    metrics_expected["precision"][4] = torch.tensor([1, 0.6666, 0.6666, 0.6666, 1, 1]).float()
    metrics_expected["map"][4] = torch.tensor([0.9166, 0.5, 0.3888, 0.2777, 1, 0.25]).float()

    metrics_expected["precision"][5] = torch.tensor([1, 1, 1, 1, 1, 1]).float()
    metrics_expected["map"][5] = torch.tensor([0.9166, 0.7, 0.5888, 0.4777, 1, 0.25]).float()

    metrics_expected["precision"][10] = metrics_expected["precision"][5]
    metrics_expected["map"][10] = metrics_expected["map"][5]

    compare_metrics(positions, labels, is_query, is_gallery, metrics_expected, top_k, reduce=False)


@pytest.mark.parametrize("max_num_labels", list(range(5, 10)))
@pytest.mark.parametrize("max_num_samples_per_label", list(range(3, 10)))
@pytest.mark.parametrize("is_query_all", [True, False])
@pytest.mark.parametrize("is_gallery_all", [True, False])
@pytest.mark.parametrize("num_attempts", [20])
def test_on_synthetic_cases(
    max_num_labels: int, max_num_samples_per_label: int, num_attempts: int, is_query_all: bool, is_gallery_all: bool
) -> None:
    top_k = (1, 2, 3, 4, 5)

    for _ in range(num_attempts):
        case = generate_retrieval_case(
            max_labels=max_num_labels,
            max_samples_per_label=max_num_samples_per_label,
            is_query_all=is_query_all,
            is_gallery_all=is_gallery_all,
            return_desired_correct_positions=True,
        )
        positions, labels, is_query, is_gallery = case

        metrics_expected = defaultdict(dict)  # type: ignore

        metrics_expected["cmc"] = {k: naive_cmc(positions, k) for k in top_k}
        metrics_expected["precision"] = {k: naive_precision(positions, k) for k in top_k}
        metrics_expected["map"] = {k: naive_map(positions, k) for k in top_k}

        compare_metrics(positions, labels, is_query, is_gallery, metrics_expected, top_k, reduce=True)


def test_validate_dataset_good_case() -> None:
    isq = np.r_[True, False, False, True, False, False]
    isg = np.r_[False, True, True, False, True, True]
    labels = np.r_[0, 0, 0, 1, 1, 1]

    mgt = calc_gt_mask(labels=labels, is_query=isq, is_gallery=isg)
    m2i = calc_mask_to_ignore(is_query=isq, is_gallery=isg)
    validate_dataset(mask_gt=mgt, mask_to_ignore=m2i)


def test_validate_dataset_bad_case() -> None:
    with pytest.raises(AssertionError):
        isq = np.r_[True, False, False, True, True]
        isg = np.r_[False, True, True, False, True]
        labels = np.r_[0, 0, 0, 1, 1]

        mgt = calc_gt_mask(labels=labels, is_query=isq, is_gallery=isg)
        m2i = calc_mask_to_ignore(is_query=isq, is_gallery=isg)
        validate_dataset(mask_gt=mgt, mask_to_ignore=m2i)


def compare_metrics(
    positions: TPositions,
    labels: torch.Tensor,
    is_query: torch.Tensor,
    is_gallery: torch.Tensor,
    metrics_expected: TMetricsDict,
    top_k: Tuple[int, ...],
    reduce: bool,
) -> None:
    distances = generate_distance_matrix(positions, labels=labels, is_query=is_query, is_gallery=is_gallery)
    mask_to_ignore = calc_mask_to_ignore(is_query=is_query, is_gallery=is_gallery)
    mask_gt = calc_gt_mask(labels=labels, is_query=is_query, is_gallery=is_gallery)

    metrics_calculated = calc_retrieval_metrics(
        distances=distances,
        mask_gt=mask_gt,
        mask_to_ignore=mask_to_ignore,
        map_top_k=top_k,
        precision_top_k=top_k,
        cmc_top_k=top_k,
        reduce=reduce,
    )

    for metric_name in metrics_expected.keys():
        for k in top_k:
            values_expected = metrics_expected[metric_name][k]
            values_calculated = metrics_calculated[metric_name][k]
            assert torch.all(torch.isclose(values_expected, values_calculated, atol=1e-4)), [metric_name, k]



from itertools import combinations, product
from random import sample
from sys import maxsize
from typing import List

from torch import Tensor

from oml.interfaces.miners import ITripletsMinerInBatch, TTripletsIds
from oml.utils.misc import find_value_ids


class AllTripletsMiner(ITripletsMinerInBatch):
    """
    This miner selects all the possible triplets for the given batch.

    """

    def __init__(self, max_output_triplets: int = maxsize):
        """
        Args:
            max_output_triplets: Number of all of the possible triplets
              in the batch can be very large, so we can limit them vis this parameter.

        """
        self._max_out_triplets = max_output_triplets

    def _sample(self, *_: Tensor, labels: List[int]) -> TTripletsIds:  # type: ignore
        """
        Args:
            labels: Labels of the samples in the batch
            *_: Note, that we ignore features argument

        Returns:
            Indices of the triplets

        """
        return get_available_triplets(labels, max_out_triplets=self._max_out_triplets)


def get_available_triplets(labels: List[int], max_out_triplets: int = maxsize) -> TTripletsIds:
    """
    For each label, the function generates all possible positive and negative pairs of triplets.
    """
    num_labels = len(labels)

    triplets = []
    for label in set(labels):
        ids_pos_cur = find_value_ids(labels, label)
        ids_neg_cur = set(range(num_labels)) - set(ids_pos_cur)

        # (l0, l1, n) and (l1, l0, n) are 2 different triplets
        # and we want both of them
        pos_pairs = list(combinations(ids_pos_cur, r=2)) + list(combinations(ids_pos_cur[::-1], r=2))

        tri = [(a, p, n) for (a, p), n in product(pos_pairs, ids_neg_cur)]
        triplets.extend(tri)

    triplets = sample(triplets, min(len(triplets), max_out_triplets))
    ids_anchor, ids_pos, ids_neg = zip(*triplets)

    return list(ids_anchor), list(ids_pos), list(ids_neg)


__all__ = ["AllTripletsMiner", "get_available_triplets"]



from typing import Any

import pytest

from oml.utils.misc import flatten_dict


@pytest.fixture()
def case1() -> Any:
    input_dict = {1: {"dog": 0.5, "cat": 0.3}, 5: {"dog": 0.7, "cat": 0.2}, 7: {10: {"dog": 0.5, "cat": 0.3}}}

    target_dict = {
        "metric/1/dog": 0.5,
        "metric/1/cat": 0.3,
        "metric/5/dog": 0.7,
        "metric/5/cat": 0.2,
        "metric/7/10/dog": 0.5,
        "metric/7/10/cat": 0.3,
    }

    return input_dict, target_dict


def test_converter(case1) -> None:  # type: ignore
    input_dict, target_dict_gt = case1

    target_dict = flatten_dict(input_dict, parent_key="metric")

    assert target_dict_gt == target_dict



import numpy as np

from oml.utils.misc import smart_sample


def test_sample_enough_items() -> None:
    """Check smart_sample in case of n_sample < len(array)."""
    for _ in range(10):
        size = np.random.randint(10, 150)
        array = list(range(size))
        n_samples = np.random.randint(1, 9)
        samples = smart_sample(array=array, k=n_samples)
        assert len(set(samples)) == n_samples
        assert len(samples) == n_samples
        assert set(samples) <= set(array)


def test_sample_not_enough_items() -> None:
    """Check smart_sample in case of n_sample > len(array)."""
    for _ in range(10):
        size = np.random.randint(2, 25)
        array = list(range(size))
        n_samples = np.random.randint(size + 1, 50)
        samples = smart_sample(array=array, k=n_samples)
        assert len(set(samples)) == size
        assert len(samples) == n_samples
        assert set(samples) == set(array)



from itertools import chain

import numpy as np
import pytest
import torch

from oml.metrics.accumulation import Accumulator

from .utils import init_ddp, run_in_ddp


@pytest.mark.parametrize("world_size", [1, 2, 3])
@pytest.mark.parametrize("device", ["cpu", "cuda"] if torch.cuda.is_available() else ["cpu"])
def test_ddp_accumulator(world_size: int, device: str) -> None:
    run_in_ddp(world_size=world_size, fn=check_ddp_accumulator, args=(device,))


def check_ddp_accumulator(rank: int, world_size: int, device: str) -> None:
    init_ddp(rank, world_size)

    value = rank + 1

    data = {
        "list": [value] * value,
        "tensor_1d": value * torch.ones(value, device=device),
        "tensor_3d": value * torch.ones((value, 2, 3), device=device),
        "numpy_1d": value * np.ones(value),
        "numpy_3d": value * np.ones((value, 4, 5)),
    }

    acc = Accumulator(keys_to_accumulate=list(data.keys()))
    acc.refresh(len(data["list"]))
    acc.update_data(data)

    synced_data = acc.sync().storage

    len_after_sync = sum(range(1, world_size + 1))

    assert len(synced_data["list"]) == len_after_sync

    assert synced_data["tensor_1d"].ndim == 1  # type: ignore
    assert synced_data["tensor_1d"].shape == (len_after_sync,)  # type: ignore

    assert synced_data["tensor_3d"].ndim == 3  # type: ignore
    assert synced_data["tensor_3d"].shape == (len_after_sync, 2, 3)  # type: ignore

    assert synced_data["numpy_1d"].ndim == 1  # type: ignore
    assert synced_data["numpy_1d"].shape == (len_after_sync,)  # type: ignore

    assert synced_data["numpy_3d"].ndim == 3  # type: ignore
    assert synced_data["numpy_3d"].shape == (len_after_sync, 4, 5)  # type: ignore

    # 'sync' doesn't guarantee ordered data, so we can only compare the data structure
    expected_list = list(chain(*[[v] * v for v in range(1, world_size + 1)]))
    assert list(sorted(synced_data["list"])) == list(sorted(expected_list))

    expected_tensor_1d = torch.cat([v * torch.ones(v) for v in range(1, world_size + 1)], dim=0)
    sorted_synced_array, _ = torch.sort(synced_data["tensor_1d"], dim=0)
    assert torch.all(torch.isclose(expected_tensor_1d, sorted_synced_array))

    expected_tensor_3d = torch.cat([v * torch.ones((v, 2, 3)) for v in range(1, world_size + 1)], dim=0)
    sorted_synced_array, _ = torch.sort(synced_data["tensor_3d"], dim=0)
    assert torch.all(torch.isclose(expected_tensor_3d, sorted_synced_array))

    expected_numpy_1d = np.concatenate([v * np.ones(v) for v in range(1, world_size + 1)], axis=0)
    sorted_synced_array = np.sort(synced_data["numpy_1d"], axis=0)
    assert np.all(np.isclose(expected_numpy_1d, sorted_synced_array))

    expected_numpy_3d = np.concatenate([v * np.ones((v, 4, 5)) for v in range(1, world_size + 1)], axis=0)
    sorted_synced_array = np.sort(synced_data["numpy_3d"], axis=0)
    assert np.all(np.isclose(expected_numpy_3d, sorted_synced_array))



import subprocess
from pathlib import Path

import pytest

from oml.const import PROJECT_ROOT


def find_code_block(file: Path, start_indicator: str, end_indicator: str) -> str:
    with open(file) as f:
        text = f.readlines()

    i = text.index(start_indicator) + 2
    j = text.index(end_indicator) - 1

    code_block = "".join(text[i:j])

    print(code_block)

    return code_block


@pytest.mark.parametrize(
    "start_indicator,end_indicator",
    [
        ("[comment]:vanilla-train-start\n", "[comment]:vanilla-train-end\n"),
        ("[comment]:vanilla-validation-start\n", "[comment]:vanilla-validation-end\n"),
        ("[comment]:lightning-start\n", "[comment]:lightning-end\n"),
        ("[comment]:lightning-ddp-start\n", "[comment]:lightning-ddp-end\n"),
        ("[comment]:checkpoint-start\n", "[comment]:checkpoint-end\n"),
    ],
)
def test_code_blocks_in_readme(start_indicator: str, end_indicator: str) -> None:
    code = find_code_block(PROJECT_ROOT / "README.md", start_indicator, end_indicator)
    tmp_fname = "tmp.py"

    with open(tmp_fname, "w") as f:
        f.write(code)

    try:
        subprocess.run(f"python {tmp_fname}", check=True, shell=True)
    finally:
        Path(tmp_fname).unlink()



from typing import List, Tuple

import pytest
import torch

from oml.functional.metrics import (
    apply_mask_to_ignore,
    calc_gt_mask,
    calc_mask_to_ignore,
)

from .synthetic import generate_distance_matrix, generate_retrieval_case

TCase = Tuple[List[List[int]], torch.Tensor, torch.Tensor, torch.Tensor]


def case_1() -> TCase:
    labels = torch.tensor([2, 2, 2, 0, 0, 0, 0, 1, 1])
    is_query = torch.tensor([True, False, True, False, True, False, True, False, True])
    is_gallery = torch.tensor([True] * len(labels))

    desired_positions = [[1, 3], [2, 4], [0, 3, 5], [5, 6, 7], [1]]

    return desired_positions, labels, is_query, is_gallery


def case_2() -> TCase:
    labels = torch.tensor([2, 2, 2, 0, 0, 0, 0, 1, 1])
    is_query = torch.tensor([True, False, True, False, True, False, True, False, True])
    is_gallery = torch.logical_not(is_query)

    desired_positions = [[2], [0], [0, 1], [2, 3], [3]]

    return desired_positions, labels, is_query, is_gallery


@pytest.mark.parametrize("case", [case_1(), case_2()])
def test_exact_cases(case: TCase) -> None:
    check_distance_matrix(case)


@pytest.mark.parametrize("max_num_labels", list(range(3, 10)))
@pytest.mark.parametrize("max_num_samples_per_label", list(range(3, 10)))
@pytest.mark.parametrize("is_query_all", [True, False])
@pytest.mark.parametrize("is_gallery_all", [True, False])
@pytest.mark.parametrize("num_attempts", [20])
def test_on_synthetic_cases(
    max_num_labels: int, max_num_samples_per_label: int, num_attempts: int, is_query_all: bool, is_gallery_all: bool
) -> None:
    for _ in range(num_attempts):
        case = generate_retrieval_case(
            max_labels=max_num_labels,
            max_samples_per_label=max_num_samples_per_label,
            is_query_all=is_query_all,
            is_gallery_all=is_gallery_all,
            return_desired_correct_positions=True,
        )
        check_distance_matrix(case)


def check_distance_matrix(case: TCase) -> None:
    desired_positions_array, labels, is_query, is_gallery = case
    distances = generate_distance_matrix(
        desired_positions_array, labels=labels, is_query=is_query, is_gallery=is_gallery
    )

    mask_to_ignore = calc_mask_to_ignore(is_query=is_query, is_gallery=is_gallery)
    mask_gt = calc_gt_mask(labels=labels, is_query=is_query, is_gallery=is_gallery)
    distances, mask_gt = apply_mask_to_ignore(distances=distances, mask_gt=mask_gt, mask_to_ignore=mask_to_ignore)

    _, ids_sorted_distance = torch.topk(distances, largest=False, k=distances.shape[1])

    for query_label, ids_sorted_for_query, desired_positions in zip(
        labels[is_query], ids_sorted_distance, desired_positions_array
    ):
        actual_positions = torch.nonzero(labels[is_gallery][ids_sorted_for_query] == query_label)
        actual_positions = actual_positions[: len(desired_positions)].squeeze(1).tolist()
        assert actual_positions == desired_positions, [actual_positions, desired_positions]



import tempfile
from functools import partial
from typing import Any, Dict, List

import pytest
import pytorch_lightning as pl
import torch
from torch import nn
from torch.optim import Adam
from torch.utils.data import DataLoader, Dataset

from oml.const import (
    EMBEDDINGS_KEY,
    INPUT_TENSORS_KEY,
    IS_GALLERY_KEY,
    IS_QUERY_KEY,
    LABELS_KEY,
)
from oml.datasets.triplet import TItem, tri_collate
from oml.lightning.callbacks.metric import MetricValCallback
from oml.losses.triplet import TripletLossPlain, TripletLossWithMiner
from oml.metrics.embeddings import EmbeddingMetrics
from oml.metrics.triplets import AccuracyOnTriplets
from oml.samplers.balance import BalanceSampler


class DummyTripletDataset(Dataset):
    def __init__(self, num_triplets: int, im_size: int):
        self.num_triplets = num_triplets
        self.im_size = im_size

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        input_tensors = torch.rand((3, 3, self.im_size, self.im_size))
        return {INPUT_TENSORS_KEY: input_tensors}

    def __len__(self) -> int:
        return self.num_triplets


class DummyRetrievalDataset(Dataset):
    def __init__(self, labels: List[int], im_size: int):
        self.labels = labels
        self.im_size = im_size

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        input_tensors = torch.rand((3, self.im_size, self.im_size))
        label = torch.tensor(self.labels[idx]).long()
        return {INPUT_TENSORS_KEY: input_tensors, LABELS_KEY: label, IS_QUERY_KEY: True, IS_GALLERY_KEY: True}

    def __len__(self) -> int:
        return len(self.labels)


class DummyCommonModule(pl.LightningModule):
    def __init__(self, im_size: int):
        super().__init__()
        self.model = nn.Sequential(
            nn.AvgPool2d(kernel_size=(im_size, im_size)), nn.Flatten(start_dim=1), nn.Linear(3, 5), nn.Linear(5, 5)
        )

    def configure_optimizers(self) -> torch.optim.Optimizer:
        return Adam(self.model.parameters(), lr=1e-4)

    def validation_step(self, batch: TItem, batch_idx: int, *dataset_idx: int) -> Dict[str, Any]:
        embeddings = self.model(batch[INPUT_TENSORS_KEY])
        return {**batch, **{EMBEDDINGS_KEY: embeddings.detach().cpu()}}


class DummyTripletModule(DummyCommonModule):
    def __init__(self, im_size: int):
        super().__init__(im_size=im_size)
        self.criterion = TripletLossPlain(margin=None)

    def training_step(self, batch_multidataloader: List[TItem], batch_idx: int) -> torch.Tensor:
        embeddings = torch.cat([self.model(batch[INPUT_TENSORS_KEY]) for batch in batch_multidataloader])
        loss = self.criterion(embeddings)
        return loss


class DummyRetrievalModule(DummyCommonModule):
    def __init__(self, im_size: int):
        super().__init__(im_size=im_size)
        self.criterion = TripletLossWithMiner(margin=None, need_logs=True)

    def training_step(self, batch_multidataloader: List[TItem], batch_idx: int) -> torch.Tensor:
        embeddings = torch.cat([self.model(batch[INPUT_TENSORS_KEY]) for batch in batch_multidataloader])
        labels = torch.cat([batch[LABELS_KEY] for batch in batch_multidataloader])
        loss = self.criterion(embeddings, labels)
        return loss


def create_triplet_dataloader(num_samples: int, im_size: int, num_workers: int) -> DataLoader:
    dataset = DummyTripletDataset(num_triplets=num_samples, im_size=im_size)
    dataloader = DataLoader(
        dataset=dataset, batch_size=num_samples // 2, num_workers=num_workers, collate_fn=tri_collate
    )
    return dataloader


def create_retrieval_dataloader(
    num_samples: int, im_size: int, n_labels: int, n_instances: int, num_workers: int
) -> DataLoader:
    assert num_samples % (n_labels * n_instances) == 0

    labels = [idx // n_instances for idx in range(num_samples)]

    dataset = DummyRetrievalDataset(labels=labels, im_size=im_size)

    sampler_retrieval = BalanceSampler(labels=labels, n_labels=n_labels, n_instances=n_instances)
    train_retrieval_loader = DataLoader(
        dataset=dataset,
        batch_sampler=sampler_retrieval,
        num_workers=num_workers,
    )
    return train_retrieval_loader


def create_triplet_callback(loader_idx: int, samples_in_getitem: int) -> MetricValCallback:
    metric = AccuracyOnTriplets(embeddings_key=EMBEDDINGS_KEY)
    metric_callback = MetricValCallback(metric=metric, loader_idx=loader_idx, samples_in_getitem=samples_in_getitem)
    return metric_callback


def create_retrieval_callback(loader_idx: int, samples_in_getitem: int) -> MetricValCallback:
    metric = EmbeddingMetrics(
        embeddings_key=EMBEDDINGS_KEY,
        labels_key=LABELS_KEY,
        is_query_key=IS_QUERY_KEY,
        is_gallery_key=IS_GALLERY_KEY,
    )
    metric_callback = MetricValCallback(metric=metric, loader_idx=loader_idx, samples_in_getitem=samples_in_getitem)
    return metric_callback


@pytest.mark.parametrize(
    "samples_in_getitem, is_error_expected, pipeline",
    [
        (1, True, "triplet"),
        (3, False, "triplet"),
        (5, True, "triplet"),
        (1, False, "retrieval"),
        (2, True, "retrieval"),
    ],
)
@pytest.mark.parametrize("num_dataloaders", [1, 2])
def test_lightning(
    samples_in_getitem: int, is_error_expected: bool, num_dataloaders: int, pipeline: str, num_workers: int
) -> None:
    num_samples = 12
    im_size = 6
    n_labels = 2
    n_instances = 3

    if pipeline == "triplet":
        create_dataloader = create_triplet_dataloader
        lightning_module = DummyTripletModule(im_size=im_size)
        create_callback = create_triplet_callback
    elif pipeline == "retrieval":
        create_dataloader = partial(create_retrieval_dataloader, n_labels=n_labels, n_instances=n_instances)
        lightning_module = DummyRetrievalModule(im_size=im_size)
        create_callback = create_retrieval_callback
    else:
        raise ValueError

    train_dataloaders = [
        create_dataloader(num_samples=num_samples, im_size=im_size, num_workers=num_workers)
        for _ in range(num_dataloaders)
    ]
    val_dataloaders = [
        create_dataloader(num_samples=num_samples, im_size=im_size, num_workers=num_workers)
        for _ in range(num_dataloaders)
    ]
    callbacks = [create_callback(loader_idx=k, samples_in_getitem=samples_in_getitem) for k in range(num_dataloaders)]

    trainer = pl.Trainer(
        default_root_dir=tempfile.gettempdir(),
        max_epochs=2,
        enable_progress_bar=False,
        num_nodes=1,
        gpus=None,
        replace_sampler_ddp=False,
        callbacks=callbacks,
        num_sanity_val_steps=0,
    )

    if is_error_expected:
        with pytest.raises(ValueError, match=callbacks[0].metric.__class__.__name__):
            trainer.fit(model=lightning_module, train_dataloaders=train_dataloaders, val_dataloaders=val_dataloaders)
            trainer.validate(model=lightning_module, dataloaders=val_dataloaders)
    else:
        trainer.fit(model=lightning_module, train_dataloaders=train_dataloaders, val_dataloaders=val_dataloaders)
        trainer.validate(model=lightning_module, dataloaders=val_dataloaders)



from abc import ABC, abstractmethod
from typing import Any, Dict

import numpy as np
from torch.utils.data import Dataset

from oml.const import INPUT_TENSORS_KEY, IS_GALLERY_KEY, IS_QUERY_KEY, LABELS_KEY
from oml.samplers.balance import BalanceSampler  # noqa


class IDatasetWithLabels(Dataset, ABC):
    """
    This is an interface for the datasets which can provide their labels.

    """

    def __getitem__(self, item: int) -> Dict[str, Any]:
        """

        Args:
            item: Idx of the sample

        Returns:
             Dictionary with the following keys:

            >>> INPUT_TENSORS_KEY
            >>> LABELS_KEY

        """
        raise NotImplementedError()

    @abstractmethod
    def get_labels(self) -> np.ndarray:
        raise NotImplementedError()


class IDatasetQueryGallery(Dataset, ABC):
    """
    This is an interface for the datasets which can provide the information on how to split
    the validation set into the two parts: query and gallery.

    """

    @abstractmethod
    def __getitem__(self, item: int) -> Dict[str, Any]:
        """
        Args:
            item: Idx of the sample

        Returns:
             Dictionary with the following keys:

            >>> INPUT_TENSORS_KEY
            >>> LABELS_KEY
            >>> IS_QUERY_KEY
            >>> IS_GALLERY_KEY

        """
        raise NotImplementedError()


__all__ = ["IDatasetWithLabels", "IDatasetQueryGallery"]



import subprocess
from collections import defaultdict
from pathlib import Path
from typing import List, Tuple

import torch
from torch import nn

from oml.const import PROJECT_ROOT

from .run_triplets_experiment_ddp import DummyModule

"""
MOTIVATION

With this experiment we want to:
1) Test patching of default dataloaders with `shuffle=False` and `shuffle=True`.
2) Test equality of models with different combinations of parameters `B = sum(Bi * N)`. Where `B` - batch size per
step which is equal for each experiment, `N` - number of devices,
`Bi` - batch size per device.
3) Loaders return the same collections of ids on each step on several epochs.

"""


# TODO: check internal `/` on widnows
exp_file = PROJECT_ROOT / "tests/test_examples/test_ddp_cases/run_triplets_experiment_ddp.py"


def test_epochs_are_equal() -> None:
    max_epochs = 2
    models = {}
    train_ids = defaultdict(list)
    val_ids = defaultdict(list)

    for experiment, (devices, len_dataset, batch_size) in enumerate(
        [(1, 120, 30), (2, 120, 15), (3, 120, 10), (2, 120, 20)]
    ):
        params = (
            f"--devices {devices}",
            f"--len_dataset {len_dataset}",
            f"--exp_num {experiment}",
            f"--batch_size {batch_size}",
            f"--max_epochs {max_epochs}",
        )
        params = " ".join(params)

        cmd = f"python {exp_file} " + params
        subprocess.run(cmd, check=True, shell=True)

        ckpt_path = DummyModule.save_path_ckpt_pattern.format(experiment=experiment)
        models[experiment] = torch.load(ckpt_path, map_location="cpu").requires_grad_(False)
        Path(ckpt_path).unlink(missing_ok=True)

        for epoch in range(max_epochs):
            train_ids_path = DummyModule.save_path_train_ids_pattern.format(experiment=experiment, epoch=epoch)
            train_ids[experiment].append(torch.load(train_ids_path))
            Path(train_ids_path).unlink(missing_ok=True)

            val_ids_path = DummyModule.save_path_val_ids_pattern.format(experiment=experiment, epoch=epoch)
            val_ids[experiment].append(torch.load(val_ids_path))
            Path(val_ids_path).unlink(missing_ok=True)

    assert train_ids[0] == train_ids[1]
    assert train_ids[0] == train_ids[2]
    assert train_ids[0] != train_ids[3]

    assert val_ids[0] == val_ids[1]
    assert val_ids[0] == val_ids[2]
    assert val_ids[0] != val_ids[3]

    eq_01, not_eq_tensors_01 = is_equal_models(models[0], models[1])
    eq_02, not_eq_tensors_02 = is_equal_models(models[0], models[2])
    eq_03, _ = is_equal_models(models[0], models[3])

    assert eq_01, not_eq_tensors_01
    assert eq_02, not_eq_tensors_02
    assert not eq_03


TORCH_EPS = 10 * torch.finfo(torch.float32).eps


def is_equal_models(model1: nn.Module, model2: nn.Module) -> Tuple[bool, List[torch.Tensor]]:
    for module1, module2 in zip(model1.modules(), model2.modules()):
        assert type(module1) == type(module2)
        if isinstance(module1, nn.Linear):
            if not torch.all(torch.isclose(module1.weight, module2.weight, atol=TORCH_EPS)):
                return False, [module1.weight, module2.weight]

    return True, []



from pathlib import Path
from typing import Any, Dict, Iterable, Optional

import torch

from oml.interfaces.models import IExtractor
from oml.models.utils import TStateDict, filter_state_dict, patch_device_and_float
from oml.models.vit.vision_transformer_clip import VisionTransformer
from oml.utils.io import download_checkpoint


class ViTCLIPExtractor(IExtractor):
    pretrained_models: Dict[str, Any] = {
        # checkpoints pretrained by OpenAI
        "openai_vitb16_224": (
            "https://openaipublic.azureedge.net/clip/models/5806e77cd80f8b59890b7e101eabd078d9fb84e6937f9e85e4ecb61988df416f/ViT-B-16.pt",
            "44c3d804ecac03d9545ac1a3adbca3a6",
        ),
        "openai_vitb32_224": (
            "https://openaipublic.azureedge.net/clip/models/40d365715913c9da98579312b702a82c18be219cc2a73407c4526f58eba950af/ViT-B-32.pt",
            "3ba34e387b24dfe590eeb1ae6a8a122b",
        ),
        "openai_vitl14_224": (
            "https://openaipublic.azureedge.net/clip/models/b8cca3fd41ae0c99ba7e8951adf17d267cdb84cd88be6f7c2e0eca1737a03836/ViT-L-14.pt",
            "096db1af569b284eb76b3881534822d9",
        ),
        "openai_vitl14_336": (
            "https://openaipublic.azureedge.net/clip/models/3035c92b350959924f9f00213499208652fc7ea050643e8b385c2dac08641f02/ViT-L-14-336px.pt",
            "b311058cae50cb10fbfa2a44231c9473",
        ),
        # checkpoints pretrained by SberbankAI
        "sber_vitb16_224": (
            "https://huggingface.co/sberbank-ai/ruclip-vit-base-patch16-224/resolve/main/pytorch_model.bin",
            "7882e07674d78c674e33cb892a68bbfc",
        ),
        "sber_vitb32_224": (
            "https://huggingface.co/sberbank-ai/ruclip-vit-base-patch32-224/resolve/main/pytorch_model.bin",
            "e2c4dab46a3cfa608bdd762973e90d32",
        ),
        "sber_vitl14_224": (
            "https://huggingface.co/sberbank-ai/ruclip-vit-large-patch14-224/resolve/main/pytorch_model.bin",
            "9b4a1cd25d15bad4ffd2ba6e34b8a67c",
        ),
    }
    jitted_weights = {"openai_vitb16_224", "openai_vitb32_224", "openai_vitl14_224", "openai_vitl14_336"}

    def __init__(
        self,
        weights: Optional[str],
        arch: str,
        normalise_features: bool = True,
        strict_load: bool = True,
    ):
        """
        Args:
            weights: Path to weights or special key for pretrained ones or ``None`` for random initialization.
             You can check available pretrained checkpoints in ``ViTCLIPExtractor.pretrained_models``.
            arch: Might be one of ``vitb16_224``, ``vitb32_224``, ``vitl14_224``, ``vitl14_336``.
            normalise_features: Set ``True`` to normalise output features
            strict_load: Whether the weights needed to be loaded strictly. Doesn't work with OpenAI's models.
        """

        super().__init__()

        self.normalize = normalise_features

        cfg = get_vit_config_by_arch(arch)
        embed_dim = cfg["embed_dim"]
        image_resolution = cfg["image_resolution"]
        layers = cfg["layers"]
        width = cfg["width"]
        patch_size = cfg["patch_size"]
        heads = cfg["heads"]

        self.visual = VisionTransformer(
            input_resolution=image_resolution,
            patch_size=patch_size,
            layers=layers,
            width=width,
            heads=heads,
            output_dim=embed_dim,
        )

        if weights is None:
            return
        if weights in self.pretrained_models:
            jitted_weights = weights in self.jitted_weights
            weights, md5 = self.pretrained_models[weights]
            weights = download_checkpoint(weights, md5)
        else:
            jitted_weights = False

        if jitted_weights:  # check if weights are jitted
            visual = torch.jit.load(Path(weights), map_location="cpu").visual
            patch_device_and_float(visual, device="cpu")
            state_dict = visual.state_dict()
        else:
            state_dict = torch.load(Path(weights), map_location="cpu")
            state_dict = state_dict.get("state_dict", state_dict)
            state_dict = take_visual_part_of_vit_clip(state_dict, needed_keys=self.visual.state_dict().keys())

        self.visual.load_state_dict(state_dict=state_dict, strict=strict_load)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        res = self.visual.forward(x)
        if self.normalize:
            res = res / torch.linalg.norm(res, 2, dim=1, keepdim=True).detach()
        return res

    @property
    def feat_dim(self) -> int:
        return self.visual.state_dict()["proj"].shape[-1]


def take_visual_part_of_vit_clip(state_dict: TStateDict, needed_keys: Iterable[str]) -> TStateDict:
    for k in list(state_dict):
        if k.startswith("visual."):
            state_dict[k.lstrip("visual")[1:]] = state_dict.pop(k)
    state_dict = filter_state_dict(state_dict, needed_keys=needed_keys)
    return state_dict


def get_vit_config_by_arch(model_name: str) -> Dict[str, Any]:
    """
    Function which returns configuration of known CLIP models.
    Args:
        model_name: One of ``vitb16_224``, ``vitb32_224``, ``vitl14_224``, ``vitl14_336``.
    """
    CLIP_MODELS = {
        "vitb16_224": {
            "embed_dim": 512,
            "image_resolution": 224,
            "layers": 12,
            "width": 768,
            "patch_size": 16,
            "heads": 12,
        },
        "vitb32_224": {
            "embed_dim": 512,
            "image_resolution": 224,
            "layers": 12,
            "width": 768,
            "patch_size": 32,
            "heads": 12,
        },
        "vitl14_224": {
            "embed_dim": 768,
            "image_resolution": 224,
            "layers": 24,
            "width": 1024,
            "patch_size": 14,
            "heads": 16,
        },
        "vitl14_336": {
            "embed_dim": 768,
            "image_resolution": 224,
            "layers": 24,
            "width": 1024,
            "patch_size": 14,
            "heads": 16,
        },
    }

    return CLIP_MODELS[model_name]


__all__ = ["ViTCLIPExtractor"]



from math import isclose

import torch

from oml.losses.triplet import TripletLossWithMiner
from oml.miners.cross_batch import TripletMinerWithMemory
from tests.test_oml.test_miners.shared_checkers import calc_n_triplets


def test() -> None:
    expand_k = 2
    miner = TripletMinerWithMemory(bank_size_in_batches=3, tri_expand_k=expand_k)
    criterion = TripletLossWithMiner(margin=0.1, miner=miner, need_logs=True)

    labels = [0] * 4 + [1] * 3 + [2] * 6 + [3] * 2
    features = torch.rand(size=(len(labels), 16))

    _ = criterion(features=features, labels=torch.tensor(labels))
    logs = criterion.last_logs

    n_tri_orig = calc_n_triplets(labels=labels)
    n_tri_bank = n_tri_orig * (expand_k - 1)
    n_tri = n_tri_orig + n_tri_bank

    wo = n_tri_orig / n_tri
    wb = n_tri_bank / n_tri

    assert isclose(wb * logs["bank_active_tri"] + wo * logs["orig_active_tri"], logs["active_tri"], abs_tol=1e-3)
    assert isclose(wb * logs["pos_dist_bank"] + wo * logs["pos_dist_orig"], logs["pos_dist"], abs_tol=1e-3)
    assert isclose(wb * logs["neg_dist_bank"] + wo * logs["neg_dist_orig"], logs["neg_dist"], abs_tol=1e-3)



class IPostprocessor:
    """
    This is a parent class for the classes which apply some postprocessing
    after the embeddings have been extracted.
    For example, we may want to apply some method of dimension reduction or
    query-reranking.

    """

    def process(self, *args, **kwargs):  # type: ignore
        raise NotImplementedError()


__all__ = ["IPostprocessor"]



from typing import Any, Dict

from oml.interfaces.models import IExtractor
from oml.models.resnet import ResnetExtractor
from oml.models.vit.clip import ViTCLIPExtractor
from oml.models.vit.vit import ViTExtractor
from oml.utils.misc import TCfg, dictconfig_to_dict

MODELS_REGISTRY = {
    "resnet": ResnetExtractor,
    "vit": ViTExtractor,
    "vit_clip": ViTCLIPExtractor,
}


def get_extractor(model_name: str, **kwargs: Dict[str, Any]) -> IExtractor:
    extractor = MODELS_REGISTRY[model_name](**kwargs)  # type: ignore
    return extractor


def get_extractor_by_cfg(cfg: TCfg) -> IExtractor:
    cfg = dictconfig_to_dict(cfg)
    extractor = get_extractor(model_name=cfg["name"], **cfg["args"])
    return extractor


__all__ = ["MODELS_REGISTRY", "get_extractor", "get_extractor_by_cfg"]



import random
from typing import Any, Callable, Type

import numpy as np
import pytest
import torch

from oml.utils.images.images_resize import (
    inverse_karesize_bboxes,
    inverse_karesize_image,
    karesize_bboxes,
    karesize_image,
)


@pytest.mark.parametrize("constructor,allclose_func", [(np.array, np.allclose), (torch.tensor, torch.allclose)])
def test_karesize_bboxes(constructor: Type, allclose_func: Callable[[Any], Any]) -> None:  # type: ignore
    min_dim = 200
    max_dim = 600
    max_bboxes = 10
    num_attempts = 100

    for _ in range(num_attempts):
        src_hw = (random.randint(min_dim, max_dim), random.randint(min_dim, max_dim))
        new_hw = (random.randint(min_dim, max_dim), random.randint(min_dim, max_dim))
        num_bboxes = random.randint(1, max_bboxes)

        left_points = [random.uniform(0, src_hw[1] - 3) for _ in range(num_bboxes)]
        right_points = [random.uniform(left + 1, src_hw[1] - 1) for left in left_points]
        top_points = [random.uniform(0, src_hw[0] - 3) for _ in range(num_bboxes)]
        bottom_points = [random.uniform(top + 1, src_hw[0] - 1) for top in top_points]

        bboxes = list(zip(left_points, top_points, right_points, bottom_points))

        raw_bboxes = constructor(bboxes)
        new_bboxes = karesize_bboxes(raw_bboxes, src_hw, new_hw)
        prev_bboxes = inverse_karesize_bboxes(new_bboxes, new_hw, src_hw)

        assert allclose_func(prev_bboxes, raw_bboxes, rtol=1e-3)  # type: ignore


@pytest.mark.parametrize("framework,ndim", [("np", 2), ("np", 3), ("torch", 2), ("torch", 2), ("torch", 4)])
def test_karesize_images(framework: str, ndim: int) -> None:
    min_dim = 200
    max_dim = 600
    num_attempts = 20
    max_extra_dims = 5

    for _ in range(num_attempts):
        src_hw = (random.randint(min_dim, max_dim), random.randint(min_dim, max_dim))
        new_hw = (random.randint(min_dim, max_dim), random.randint(min_dim, max_dim))

        src_shape = src_hw
        new_shape = new_hw

        for _ in range(ndim - 2):
            extra_dim = random.randint(1, max_extra_dims)
            if framework == "np":
                src_shape = (*src_shape, extra_dim)  # type: ignore
                new_shape = (*new_shape, extra_dim)  # type: ignore
            else:
                src_shape = (extra_dim, *src_shape)  # type: ignore
                new_shape = (extra_dim, *new_shape)  # type: ignore

        if framework == "np":
            raw_image = np.random.random(src_shape)
        else:
            raw_image = torch.randn(src_shape)

        new_image = karesize_image(raw_image, new_hw)
        assert new_image.shape == new_shape
        prev_image = inverse_karesize_image(new_image, src_hw)
        assert prev_image.shape == src_shape
